/****************************************************************************
 *
 *   Copyright (C) 2003, 2004 Wolfgang G. Reissnegger, All Rights Reserved
 *
 *   gnagflow@users.sourceforge.net                                    
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the license, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public
 *   License along with this program;  if not, write to the
 *       Free Software Foundation, Inc.,
 *       59 Temple Place - Suite 330,
 *       Boston, MA 02111-1307
 *
 ****************************************************************************/

#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <ctype.h>

#include <assert.h>

#include "playlist.h"

#include "util.h"
#include "fileutil.h"

/******************************************************************************/

// Global file list. This list will be scanned once at application
// startup and then never be changed / deleted.
//
static list_t *file_list = NULL;

// CD file list. Will be scanned if a CD is played.
//
static list_t *file_list_cd = NULL;

/******************************************************************************
 *
 * has_suffix
 *
 ******************************************************************************/

static int has_suffix(const char *suffix, const char *fname)
{
	int flen;
	int slen;

	assert(NULL != suffix);
	assert(NULL != fname);

	flen = strlen(fname);
	slen = strlen(suffix);

	if (0 == slen)
		return FALSE;

	return ((flen > slen) && (0 == strcmp(&fname[flen - slen], suffix)));
}


/*****************************************************************************
 *
 * scan_dir
 *
 * dir -    directory to scan in
 *
 ****************************************************************************/

static void scan_dir(list_t ** list, const char *dir, int recursive)
{
	DIR *dirlst;
	struct dirent *entry;

	assert(NULL != list);
	assert(NULL != dir);

	dirlst = opendir(dir);
	if (!dirlst)
	{
		return;
	}

	while (NULL != (entry = readdir(dirlst)))
	{
		file_entry_t *fentry;

		// Ignore entries beginning with a dot
		//
		if ('.' == *entry->d_name)
		{
			continue;
		}

		fentry = file_entry_new(NULL);
		assert(NULL != fentry);

		// file_entry_new(NULL) initialized fentry->type to
		// FILE_ENTRY_MENU. Let's set it right.
		//
		fentry->type = FILE_ENTRY_FILE;

		// Add the new file entry
		//
		sprintf(fentry->fullpath, "%s/%s", dir, entry->d_name);
		strcpy(fentry->name, entry->d_name);

		*list = list_add(*list, fentry);

		// scan recursive, if requested
		//
		if (recursive && (FILE_TYPE_DIR == file_type(fentry)))
		{
			scan_dir(list, fentry->fullpath, TRUE);
		}
	}
	closedir(dirlst);
}


/******************************************************************************/

static int is_dir(const char *path)
{
#if 0
	struct stat statbuf;

	return (0 == lstat(path, &statbuf)) && S_ISDIR(statbuf.st_mode);
#endif

	// To speed things up, we avoid calling lstat here (this is really bad
	// for systems that mount their mp3 directory over NFS). We just assume
	// that the given pathname refers to a directory, if it does not have a
	// suffix in the last 4 characters.
	//
	// UPDATE:
	//   OK, this hack bit me in the a%& when I added support for .flac
	//   files. Unfortunately flac spells with 4 letters which :-(
	//   Let's change the hack...
	//
	size_t len = strlen(path);

	// Old hack
	//
#if 0
	return (len < 4) || (NULL == strchr(&path[len - 4], '.'));
#endif

	// New hack
	//
	// We assume that all suffixes have either 3 or 4 characters.
	// Therefore, the '.' must be located either at len-4 or len-5.
	//
	if ((len >= 4) && ('.' == path[len -4]))
	{
		return FALSE;	// is 3 char suffix
	}
	if ((len >= 5) && ('.' == path[len -5]))
	{
		return FALSE;	// 4 char suffix
	}
	// No suffix. We assume it's a DIR.
	//
	return TRUE;
}


/******************************************************************************/

int file_type(const file_entry_t * entry)
{
	assert(NULL != entry);

	const char *filename;

	// This might be a menu entry, return immediately, if so...
	//
	if (FILE_ENTRY_MENU == entry->type)
	{
		return FILE_TYPE_UNKNOWN;
	}

	// Otherwise, check for known file types.
	//
	filename = entry->fullpath;

	if (is_dir(filename))
	{
		return FILE_TYPE_DIR;
	}
	else if (has_suffix("mp3", filename) || has_suffix("MP3", filename))
	{
		return FILE_TYPE_MP3;
	}
	else if (has_suffix("ogg", filename) || has_suffix("OGG", filename))
	{
		return FILE_TYPE_OGG;
	}
	else if (has_suffix("flac", filename) || has_suffix("FLAC", filename))
	{
		return FILE_TYPE_FLAC;
	}
	else if (has_suffix("url", filename) || has_suffix("URL", filename))
	{
		return FILE_TYPE_URL;
	}
	else if (has_suffix("pls", filename))
	{
		return FILE_TYPE_PLAYLIST;
	}

	return FILE_TYPE_UNKNOWN;
}


/******************************************************************************/

const char *url_from_file(const char *fullpath)
{
	// NOTE:
	//   This function is not re-entrant!
	//
	static char line[MAX_URL_LEN];

	FILE *fp;
	char *p;

	assert(NULL != fullpath);

	// Terminate the buffer in case the following code fails.
	//
	*line = '\0';

	fp = fopen(fullpath, "r");
	if (fp)
	{
		if (NULL != (p = fgets(line, sizeof(line), fp)))
		{
			// remove the \n at the end, if it exists
			//
			if ('\n' == p[strlen(p) - 1])
				p[strlen(p) - 1] = '\0';
		}
		fclose(fp);
	}

	return line;
}


/******************************************************************************/

static const char *name_of(const char *fullpath)
{
	const char *p;

	assert(NULL != fullpath);

	p = fullpath + strlen(fullpath);

	while ((p > fullpath) && ('/' != *(p - 1)))
		p--;

	return p;
}


/******************************************************************************/

char *get_path(char *target, int len, const char *fullpath)
{
	const char *name_start;

	name_start = name_of(fullpath);

	// check, if it is a root path
	//
	if (1 == (name_start - fullpath))
	{
		strncpy(target, fullpath, len);
		return target;
	}

	if ((name_start - fullpath) > len - 1)
	{
		return NULL;
	}

	if (name_start > fullpath)
	{
		// copy all until slash
		//
		memcpy(target, fullpath, name_start - fullpath - 1);

		// set terminating zero
		//
		target[name_start - fullpath - 1] = '\0';
	}
	return target;
}


/******************************************************************************/

file_entry_t *file_entry_new(const char *fullpath)
{
	file_entry_t *entry;

	entry = (file_entry_t *) malloc(sizeof(file_entry_t));
	assert(NULL != entry);

	if (fullpath)
	{
		save_strcpy(entry->fullpath, fullpath);
		save_strcpy(entry->name, name_of(fullpath));
		entry->type = FILE_ENTRY_FILE;
	}
	else
	{
		entry->fullpath[0] = '\0';
		entry->name[0] = '\0';
		entry->type = FILE_ENTRY_MENU;
	}
	entry->marked = 0;

	return entry;
}


/******************************************************************************/

static int file_entry_compare(const void *arg1, const void *arg2)
{
	file_entry_t *const *e1 = arg1;
	file_entry_t *const *e2 = arg2;

	return strcmp((*e1)->fullpath, (*e2)->fullpath);
}


list_t *file_entry_list_sort(list_t *list)
{
	file_entry_t **array;
	int idx;
	int len;
	list_t *iter;

	// anything in list?
	//
	if (!list)	return NULL;

	// alloc the array for the menu elements
	//
	array = malloc(list_len(list) * sizeof(file_entry_t *));

	// fill the array
	//
	for (len = 0, iter = list_first(list); iter; iter = list_next(iter), len++)
	{
		array[len] = (file_entry_t *) list_data(iter);
	}

	// Sort the list
	//
	qsort(array, len, sizeof(file_entry_t *), file_entry_compare);

	// Delete the old list and create a new one with sorted entries
	//
	// NOTE: This does NOT delete the file entry objects, only the
	// list nodes.
	//
	list_free(list);
	list = NULL;

	for (idx = 0; idx < len; idx++)
	{
		list = list_add(list, array[idx]);
	}

	free(array);

	return list;
}


/******************************************************************************/

static void file_entry_list_free(list_t *list)
{
	list_t *iter;

	for (iter = list_first(list); iter; iter = list_next(iter))
	{
		if (list_data(iter))
			free(list_data(iter));
	}
	list_free(list);
}


/******************************************************************************
 *
 * scan_file_list_path / recursive
 *
 ******************************************************************************/

static list_t *_scan_file_list_path(const char *path, int recursive)
{
	list_t *list = NULL;
	list_t *iter;

	// First search the HD file list
	//
	for (iter = list_first(file_list); iter; iter = list_next(iter))
	{
		char buf[MAX_PATH_LEN];
		char *p;
		file_entry_t *entry = list_data(iter);

		p = get_path(buf, sizeof(buf), entry->fullpath);
		assert(NULL != p);

		// If we are searching recursively, we compare only substring
		// of p against the search path -> All entries that START with
		// the given path string will match.
		//
		// Otherwise, we are looking for an exact match of the
		// pathnames.
		//
		if (recursive)
		{
			if (0 == strncmp(path, p, strlen(path)))
			{
				list = list_add(list, entry);
			}
		}
		else
		{
			if (0 == strcmp(path, p))
			{
				list = list_add(list, entry);
			}
		}
	}

	// Then do it again for the CD file list
	//
	for (iter = list_first(file_list_cd); iter; iter = list_next(iter))
	{
		char buf[MAX_PATH_LEN];
		char *p;
		file_entry_t *entry = list_data(iter);

		p = get_path(buf, sizeof(buf), entry->fullpath);
		assert(NULL != p);

		if (0 == strcmp(path, p))
		{
			list = list_add(list, entry);
		}
	}

	return list;
}


/******************************************************************************/

list_t *scan_file_list_path(const char *path)
{
	return _scan_file_list_path(path, 0);
}


/******************************************************************************/

list_t *scan_file_list_path_recursive(const char *path)
{
	return _scan_file_list_path(path, 1);
}


/******************************************************************************/

//
// Case insensitive search. Like strstr() but case insensitive.
//
static char *istrstr(char *haystack, const char *needle)
{
	char p1[MAX_PATH_LEN];
	char p2[MAX_PATH_LEN];
	char *pp1 = p1;
	char *pp2 = p2;
	char *p;

	assert(NULL != haystack);
	assert(NULL != needle);
	assert(strlen(haystack) < MAX_PATH_LEN);
	assert(strlen(needle) < MAX_PATH_LEN);

	// We need to preserve haystack to calculate the offset
	// in the return statement. needle can be changed, though.
	//
	p = haystack;
	while (*p)
		*pp1++ = toupper(*p++);
	*pp1 = '\0';
	while (*needle)
		*pp2++ = toupper(*needle++);
	*pp2 = '\0';

	p = strstr(p1, p2);

	return p ? haystack + (p - p1) : NULL;
}


/******************************************************************************/

list_t *file_entry_list_search(list_t *haystack, const char *needle)
{
	list_t *iter;
	list_t *result = NULL;

	assert(NULL != needle);

	// Search the HD file list
	//
	for (iter = list_first(haystack); iter; iter = list_next(iter))
	{
		file_entry_t *fentry = list_data(iter);

		if (istrstr(fentry->fullpath, needle))
		{
			result = list_add(result, fentry);
		}
	}
	return result;
}


/******************************************************************************/

list_t *file_entry_list_search_all(const char *needle)
{
	list_t *result;

	assert(NULL != needle);

	result = file_entry_list_search(file_list, needle);
	result = list_concat(result, file_entry_list_search(file_list_cd, needle));

	return result;
}


/******************************************************************************
 *
 * file_entry_list_init
 *
 ******************************************************************************/

void file_entry_list_init()
{
	// Check, if we already have a file_list in which case we have to free
	// it.
	//
	if (file_list)
	{
		file_entry_list_free(file_list);
	}


	// Check, if we have a saved filelist. If yes, read it. If no, scan the
	// whole mp3 directory tree.
	//
	file_list = playlist_read_file(AUDIOBOX_PLAYLIST_CACHE_FILE);

	if (!file_list)
	{
		scan_dir(&file_list, DIR_MP3, TRUE);
		playlist_write_file(file_list, AUDIOBOX_PLAYLIST_CACHE_FILE);
	}
}


/******************************************************************************
 *
 * file_entry_list_cd_init
 *
 ******************************************************************************/

void file_entry_list_cd_init()
{
	// Remove old list before scanning new CD
	//
	file_entry_list_cd_deinit();
	scan_dir(&file_list_cd, DIR_CDROM, TRUE);
}


/******************************************************************************
 *
 * file_entry_list_usb_init
 *
 ******************************************************************************/

void file_entry_list_usb_init()
{
	// Remove old list before scanning new CD
	//
	file_entry_list_cd_deinit();
	scan_dir(&file_list_cd, DIR_USB, TRUE);
}


/******************************************************************************
 *
 * file_entry_list_deinit
 *
 ******************************************************************************/

void file_entry_list_deinit()
{
	if (file_list)
	{
		playlist_write_file(file_list, AUDIOBOX_PLAYLIST_CACHE_FILE);

		file_entry_list_free(file_list);
		file_list = NULL;
	}
}


/******************************************************************************
 *
 * file_entry_list_cd_deinit
 *
 ******************************************************************************/

void file_entry_list_cd_deinit()
{
	file_entry_list_free(file_list_cd);
	file_list_cd = NULL;
}
