/****************************************************************************
 *
 *   Copyright (C) 2003, 2004 Wolfgang G. Reissnegger, All Rights Reserved
 *
 *   gnagflow@users.sourceforge.net                                    
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the license, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public
 *   License along with this program;  if not, write to the
 *       Free Software Foundation, Inc.,
 *       59 Temple Place - Suite 330,
 *       Boston, MA 02111-1307
 *
 ****************************************************************************/

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>

#include "search.h"
#include "keytab.h"
#include "song.h"
#include "player.h"
#include "playlist.h"
#include "info.h"
#include "util.h"

#include "menu.h"

/******************************************************************************/

typedef struct menu_d
{
	// Screen base 'class' object
	//
	screen_t parent;

	// display context object
	//
	menu_display_context_t ctx;

#define m_ctx               ((menu_t *) screen)->ctx
#define m_start             ((menu_t *) screen)->ctx.start
#define m_selection         ((menu_t *) screen)->ctx.selection
#define m_view_fullpath     ((menu_t *) screen)->ctx.view_fullpath
#define m_left_margin       ((menu_t *) screen)->ctx.left_margin
}
menu_t;


/******************************************************************************/

static void menu_delete(screen_t * screen)
{
	assert(screen);

	list_free(m_start);

	free(screen);
}


/******************************************************************************/

static void menu_select_prev(screen_t * screen, int num)
{
	while (num--)
	{
		if (m_selection > 0)
		{
			m_selection--;
		}
		// Wrap around for song list menus...
		//
		else if (TYPE_SONGLIST_MENU == screen->type)
		{
			m_selection = list_len(m_start) -1;
		}
	}
}


/******************************************************************************/

static void menu_select_next(screen_t * screen, int num)
{
	while (num--)
	{
		// Check, if a next element exists
		//
		if (m_selection < list_len(m_start) - 1)
		{
			m_selection++;
		}
		// Wrap around if this is a song list screen
		//
		else if (TYPE_SONGLIST_MENU == screen->type)
		{
			m_selection = 0;
		}
	}
}


/******************************************************************************/

// menu_play_song_list was added to accomodate the KEY_SONG_PLAY_SUBDIR case.
// The standard handler can not handle this case. Actually,the handler system
// is kind of broken and does not support enough flexibility.
//
// For legacy reasons menu_play_song wraps the more general menu_play_song_list,
//
static void menu_play_song_list(list_t *songlist, int current)
{
	screen_t *song;

	assert(songlist);

	// If a song is playing, suspend the current song.
	//
	if (player_is_playing() && song_active_screen())
	{
		song_suspend(song_active_screen());
	}

	// Create a new song screen with the song list
	//
	song = song_new();
	song_set_file_list(song, list_copy(songlist), current);

	display_push(song);
	song_play(song);
}


/******************************************************************************/

static void menu_play_song(void *data)
{
	screen_t *screen = (screen_t *) data;
	menu_play_song_list(m_start, m_selection);
}


/******************************************************************************/

static void menu_scan_selected_dir(void *data)
{
	screen_t *screen = (screen_t *) data;
	file_entry_t *entry = list_data(list_nth(m_start, m_selection));

	display_push(menu_new_from_dir(entry->fullpath));
}


/******************************************************************************/

static void playlist_read()
{
	screen_t *screen = display_current_screen();
	file_entry_t *entry = list_data(list_nth(m_start, m_selection));
	list_t *filelist = playlist_read_file(entry->fullpath);

	display_push(menu_new_from_filelist(filelist));

	list_free(filelist);
}


/******************************************************************************/

static void menu_select_dir_of_current_selection(screen_t * screen)
{
	file_entry_t *fentry = list_data(list_nth(m_start, m_selection));
	list_t *filelist;
	char buf[MAX_PATH_LEN];
	char *path;


	// Get the file list of the dir
	//
	path = get_path(buf, sizeof(buf), fentry->fullpath);
	assert(NULL != path);

	filelist = scan_file_list_path(path);


	// Push a new menu screen with the new file list
	//
	display_push(menu_new_from_filelist(filelist));
}


/******************************************************************************/

static int menu_handle_key(screen_t * screen, int key)
{
	assert(screen);

	switch (key)
	{
	case KEY_DOWN:
		menu_select_next(screen, 1);
		break;
	case KEY_UP:
		menu_select_prev(screen, 1);
		break;
	case KEY_DOWN_10:
		menu_select_next(screen, 10);
		break;
	case KEY_UP_10:
		menu_select_prev(screen, 10);
		break;

	case KEY_RIGHT:
		m_left_margin++;
		break;

	case KEY_LEFT:
		if (m_left_margin > 0)
		{
			m_left_margin--;
		}
		break;

	case KEY_SELECT:
		{
			file_entry_t *entry;

			entry = list_data(list_nth(m_start, m_selection));

			if (entry && entry->handler)
			{
				(entry->handler) (screen);
			}
		}
		break;

	case KEY_SONG_PLAY_SUBDIR:
		{
			file_entry_t	*entry;
			list_t		*fentry_list;

			entry	    = list_data(list_nth(m_start, m_selection));

			fentry_list = scan_file_list_path_recursive(entry->fullpath);

			if (fentry_list)
			{
				menu_play_song_list(fentry_list, 0);
			}
		}
		break;

	case KEY_MARK:
		{
			file_entry_t *entry;

			entry = list_data(list_nth(m_start, m_selection));

			if (entry)
			{
				entry->marked = TRUE;
			}
		}
		break;

	case KEY_UNMARK:
		{
			file_entry_t *entry;

			entry = list_data(list_nth(m_start, m_selection));

			if (entry)
			{
				entry->marked = FALSE;
			}
		}
		break;

	case ' ':		// space == toggle view mode between
		// song name only and full path;
	case '\t':		// so does the TAB key
		m_view_fullpath = !m_view_fullpath;
		break;

	case '.':
		// Set menu list to all files in the same
		// dir as currently selected file
		//
		menu_select_dir_of_current_selection(screen);
		break;

	case KEY_INSERT:
		// Add the currently selected song to the play-
		// list of the currently active song screen
		//
		if (song_active_screen())
		{
			file_entry_t *entry;

			entry = list_data(list_nth(m_start, m_selection));
			song_file_list_add(song_active_screen(), entry);

			display_push(info_new("Song added to list...", 2000));
		}
		break;

	case KEY_SEARCH:
		// If we start a search within a menu, we use
		// the current file list as search pool.
		//
		display_push(search_new(list_copy(m_start)));
		break;

	case KEY_BACKSPACE:
	case KEY_ESC:
		display_pop();
		break;
	}

	// do a first letter search on the list
	//
	key = toupper(key);
	if (((key >= 'A') && (key <= 'Z')) || ((key >= '0') && (key <= '9')))
	{
		list_t *iter;
		file_entry_t *entry = NULL;

		for (iter = list_first(m_start); iter; iter = list_next(iter))
		{
			entry = list_data(iter);

			if (key <= toupper(entry->name[0]))
			{
				break;
			}
		}
		m_selection = list_index_of(m_start, entry);
	}

	return KEY_HANDLED;
}


/******************************************************************************/

screen_t *menu_new()
{
	screen_t *screen;

	// NOTE: We actually allocate a menu_t object, but we handle it with
	// its parent 'class' (screen_t) pointer.
	//
	screen = (screen_t *) malloc(sizeof(menu_t));
	assert(NULL != screen);

	bzero(screen, sizeof(menu_t));
	screen->type		= TYPE_MENU;
	screen->update		= NULL;	// not needed for menu
	screen->delete		= menu_delete;
	screen->handle_key	= menu_handle_key;
	screen->ctx		= &m_ctx;

	return screen;
}


/******************************************************************************/

screen_t *menu_new_from_filelist(list_t *fentry_list)
{
	list_t		*iter;
	screen_t	*screen = menu_new();

	// At this point we have to set the screen type to TYPE_SONGLIST_MENU,
	// as the menu navigation code relies on it and implements different
	// behaviour for regular menus and song list menus.
	//
	// At some point in the future it might be a good idea to create a new
	// filelist menu screen...
	//
	screen->type = TYPE_SONGLIST_MENU;

	// Go through the list and set the appropriate handler
	// for each file entry.
	//
	for (iter = list_first(fentry_list); iter; iter = list_next(iter))
	{
		file_entry_t *fentry = list_data(iter);

		switch (file_type(fentry))
		{
		case FILE_TYPE_DIR:
			fentry->handler = menu_scan_selected_dir;
			break;

		case FILE_TYPE_PLAYLIST:
			fentry->handler = playlist_read;
			break;

		default:
			fentry->handler = menu_play_song;
		}
	}

	m_start = file_entry_list_sort(fentry_list);

	return screen;
}


/******************************************************************************/

screen_t *menu_new_from_dir(const char *dir)
{
	list_t *fentry_list;
	screen_t *menu;

	assert(NULL != dir);

	fentry_list = scan_file_list_path(dir);

	menu = menu_new_from_filelist(fentry_list);

	return menu;
}


/******************************************************************************/

file_entry_t *menu_add(screen_t * screen, const char *name, void (*h) (void *))
{
	file_entry_t *entry;

	entry = file_entry_new(NULL);
	assert(NULL != entry);
	save_strcpy(entry->name, name);

	entry->handler = h;

	m_start = list_add(m_start, entry);

	return entry;
}
